
#include "bitmap.h"

static void TO_LE16(uint8_t *dst, uint16_t value) {
	for (int i = 0; i < 2; ++i) {
		dst[i] = value & 255;
		value >>= 8;
	}
}

static void TO_LE32(uint8_t *dst, uint32_t value) {
	for (int i = 0; i < 4; ++i) {
		dst[i] = value & 255;
		value >>= 8;
	}
}

static void convert_amiga_color(uint8_t *p, uint16_t color) {
	const int r = (color >> 8) & 15;
	const int g = (color >> 4) & 15;
	const int b =  color       & 15;
	p[0] = (b << 4) | b;
	p[1] = (g << 4) | g;
	p[2] = (r << 4) | r;
	p[3] = 0;
}

uint8_t *bitmap_from_pal_data(const uint8_t *src, int w, int h, const uint16_t *amiga_pal, int transparent_color) {
	static const int FILE_HEADER_SIZE = 14;
	static const int INFO_HEADER_SIZE = 40;
	const int aligned_width = (w + 3) & ~3;
	const int image_size = aligned_width * h;
	const int buffer_size = FILE_HEADER_SIZE + INFO_HEADER_SIZE + 256 * 4 + image_size;
	uint8_t *buffer = (uint8_t *)calloc(buffer_size, 1);
	if (buffer) {
		int offset = 0;

		// file header
		TO_LE16(buffer + offset, 0x4D42); offset += 2;
		TO_LE32(buffer + offset, buffer_size); offset += 4;
		TO_LE16(buffer + offset, 0); offset += 2; // reserved1
		TO_LE16(buffer + offset, 0); offset += 2; // reserved2
		TO_LE32(buffer + offset, FILE_HEADER_SIZE + INFO_HEADER_SIZE + 256 * 4); offset += 4;

		// info header
		TO_LE32(buffer + offset, INFO_HEADER_SIZE); offset += 4;
		TO_LE32(buffer + offset, w); offset += 4;
		TO_LE32(buffer + offset, h); offset += 4;
		TO_LE16(buffer + offset, 1); offset += 2; // planes
		TO_LE16(buffer + offset, 8); offset += 2; // bit_count
		TO_LE32(buffer + offset, 0); offset += 4; // compression
		TO_LE32(buffer + offset, image_size); offset += 4; // size_image
		TO_LE32(buffer + offset, 0); offset += 4; // x_pels_per_meter
		TO_LE32(buffer + offset, 0); offset += 4; // y_pels_per_meter
		TO_LE32(buffer + offset, 0); offset += 4; // num_colors_used
		TO_LE32(buffer + offset, 0); offset += 4; // num_colors_important

		// palette data
		for (int i = 1; i < 16; ++i) {
			convert_amiga_color(buffer + offset + i * 4, amiga_pal[i]);
		}
		if (transparent_color != -1) {
			const int toffset = offset + transparent_color * 4;
			buffer[toffset] = 255; buffer[toffset + 1] = 0; buffer[toffset + 2] = 255; // color_key
		}
		offset += 256 * 4;

		// bitmap data
		for (int i = 0; i < h; ++i) {
			memcpy(buffer + offset, src + (h - 1 - i) * w, w);
			offset += aligned_width;
		}
	}
	return buffer;
}
