
#include "decode.h"

void decode_image_ega(uint8_t *dst, int dst_pitch, const uint8_t *src, int w, int h) {
	const int size = (w + 7) / 8;
	for (int y = 0; y < h; ++y) {
		for (int i = 0; i < size; ++i) {
			for (int b = 0; b < 8; ++b) {
				const int mask = 1 << (7 - b);
				uint8_t color = 0;
				for (int p = 0; p < 4; ++p) {
					if (src[p * size + i] & mask) {
						color |= 1 << p;
					}
				}
				dst[y * dst_pitch + i * 8 + b] = color;
			}
		}
		src += size * 4;
	}
}

void decode_sprite_ega(uint8_t *dst, int dst_pitch, const uint8_t *src, int w, int h) {
	uint8_t data[4];
	for (int y = 0; y < h; ++y) {
		for (int i = 0; i < (w + 7) / 8; ++i) {
			memcpy(data, src, 4);
			src += 4;
			for (int b = 0; b < 8; ++b) {
				const int mask = 1 << (7 - b);
				uint8_t color = 0;
				for (int p = 0; p < 4; ++p) {
					if (data[p] & mask) {
						color |= 1 << p;
					}
				}
				dst[y * dst_pitch + i * 8 + b] = color;
			}
		}
	}
}

void decode_tile(uint8_t *dst, int dst_pitch, const uint8_t *src) {
	uint16_t data[4];
	for (int y = 0; y < 16; ++y) {
		for (int p = 0; p < 4; ++p) {
			data[p] = READ_BE_UINT16(src);
			src += 2;
		}
		for (int b = 0; b < 16; ++b) {
			const int mask = 1 << (15 - b);
			uint8_t color = 0;
			for (int p = 0; p < 4; ++p) {
				if (data[p] & mask) {
					color |= 1 << p;
				}
			}
			dst[y * dst_pitch + b] = color;
		}
	}
}
