
#ifndef GAME_H__
#define GAME_H__

#include "util.h"
#include "intern.h"
#include "resource.h"

#define GAME_ANIM_NICKY_DX 4
#define GAME_ANIM_NICKY_DY 4

#define GAME_LEVELS__v3 5

#define GAME_DIR_RIGHT 1
#define GAME_DIR_LEFT  2
#define GAME_DIR_UP    4
#define GAME_DIR_DOWN  8
#define GAME_DIR_VERT  (GAME_DIR_UP | GAME_DIR_DOWN)
#define GAME_DIR_HORIZ (GAME_DIR_RIGHT | GAME_DIR_LEFT)
#define GAME_DIR_ANY   (GAME_DIR_RIGHT | GAME_DIR_LEFT | GAME_DIR_UP | GAME_DIR_DOWN)

#define GAME_INV_LIFES 0
#define GAME_INV_KEYS  1
#define GAME_INV_BULLMODES  2
#define GAME_INV_HURRICANES 3

#define GAME_LIFEBAR_STEPS 36

void game_draw_object(object_state_t *os);
uint8_t game_set_decor_tile_flags4_3_2_0(uint8_t b, int offset);
void game_set_decor_tile_flag0(uint8_t b, int offset);
uint8_t game_get_decor_tile_flag3(int offset);
void game_set_next_carlos_anim_frame();
void game_init_level_start_screen();
void game_init_ref_ref(anim_data_t *ad, uint16_t flags);
void game_init_objects_from_positref(const uint8_t *posit_ref, uint16_t objects_offset);
void game_execute_logic_op();
void game_update_anim_helper();
void	game_change_decor_tile_map_line(const object_state_t *os, const anim_data_t *ad);
void game_adjust_tilemap();
void game_play_anim_data_sfx(const anim_data_t *ad);
extern void	game_set_object_to_initial_pos(object_state_t *os, anim_data_t *ad);
void game_setup_monster(int flag);
extern void	game_init_object_from_anim_data(object_state_t *os, anim_data_t *ad);
anim_data_t *game_init_object_from_anim3(object_state_t *os, anim_data_t *ad);
void game_handle_projectiles_collisions_helper2(object_state_t *os, int _bp, int count);
anim_data_t *	game_reset_object_to_anim_pos(object_state_t *os, anim_data_t *ad);
void game_logic_op_helper2(object_state_t *_di, anim_data_t *_si, anim_data_t *ad, uint16_t move);
anim_data_t *game_logic_op2(object_state_t *os, anim_data_t *ad);
anim_data_t *game_logic_op(object_state_t *os, anim_data_t *ad);
void game_collides_op(object_state_t *os, anim_data_t *ad);
void game_collides_op_oscar(object_state_t *os, anim_data_t *ad);
int game_get_decor_offset(int x, int y);
int game_get_nicky_offset(int x, int y);
int game_test_object_bounding_box(object_state_t *os, anim_data_t *ad, int x1, int y1, int x2, int y2);
int game_test_object_on_screen(object_state_t *os, anim_data_t *ad);
void game_collide_bonus(int offset);
uint16_t game_get_decor_tile_flag7(int offset);

#define OBJECTS_SIZE 349

struct game_t {
	int current_level;
	player_state_t player_state;
	oscar_state_t _oscar_state;
	game_state_t game_state;
	decor_state_t decor_state;
	boss_state_t boss_state;
	int player_pos_x, player_pos_y;
	int cycle_counter, cycle_counter2;
	int bonus_count;
	int cycle_score;
	object_state_t objects_table[OBJECTS_SIZE];
	object_state_t *objects_table_ptr1;
	object_state_t *objects_table_ptr2;
	object_state_t *objects_table_ptr7;
	object_state_t *objects_table_ptr8;
	object_state_t *objects_table_ptr9;
	struct point_t _oscar_pos_table[20];
	anim_data_t *ref_ref_1, *ref_ref_21;
	int unk_ref_index;
	object_state_t *colliding_boss_object;
	int moving_decor_flag;
	int boss_logic_op_flag;
	int hurricanes_count;
	int carlos_bullmode_flag;
	int carlos_direction;
	int carlos_offset_dy;
	int test_object_bounding_box_offset;
	int test_object_bounding_box_counter;
	int test_object_bounding_box_result;
	int execute_op_logic_helper1_flag;
	int monster_fight_flag;
	int restart_level_flag;
	int new_level_flag;
	int quit_level_flag;
	int op_colliding_mask;
	uint16_t nicky_anim_flags1;
	int update_nicky_anim_flag;
	int update_decor_counter;
	const uint16_t *carlos_move_offsets_ptr;
	int bounding_box_x1, bounding_box_x2, bounding_box_y1, bounding_box_y2;
	object_state_t *objects_list_cur, *objects_list_head, *objects_list_last;
	int ret_flag;
	int life_bar_flag;
	int screen_img;
};

extern struct game_t g_game GCC_EXT;

extern void	game_init();
extern void	game_init_level(int level);
extern void	game_fini_level();
extern void	game_init_carlos();
extern void	game_update_life_bar();
extern void	game_init_player_state();
extern void 	game_init_oscar_state();
extern void 	game_draw_oscar_object();
extern void	game_update_oscar_object();
extern void 	game_fix_player_pos_tilemap_offset();
extern void 	game_update_object_tilemap(object_state_t *os);
extern void	game_shake_screen();
extern void	game_update_life_bar_monster();
extern void	game_set_state(int state);
extern void	game_run();

#endif // GAME_H__
