
#ifndef INTERN_H__
#define INTERN_H__

#include "util.h"

typedef struct options_t {
	int start_level;
	int start_pos_x;
	int start_pos_y;
} options_t;

typedef struct point_t {
	int16_t x;
	int16_t y;
} point_t;

typedef struct anim_data_t {
	int8_t unk0;
	int16_t lifes;
	int8_t cycles;
	int8_t unk3;
	uint8_t unk4;
	uint8_t unk5;
	uint8_t unk6;
	uint8_t init_sprite_num;
	uint8_t colliding_opcode;
	uint8_t logic_opcode;
	int8_t sound_num;
	uint8_t rnd;
	uint16_t sprite_num;
	uint16_t sprite_flags;
	uint16_t default_sprite_num;
	uint16_t default_sprite_flags;
	uint16_t anim_w;
	uint16_t anim_h;
	uint16_t score;
	int16_t bounding_box_x1;
	int16_t bounding_box_x2;
	int16_t bounding_box_y1;
	int16_t bounding_box_y2;
	const uint8_t *move_data_ptr;
	int16_t distance_dx;
	int16_t distance_dy;
	struct anim_data_t *anim_data1_ptr;
	struct anim_data_t *anim_data2_ptr;
	struct anim_data_t *anim_data3_ptr;
	int16_t dx;
	int16_t dy;
	struct anim_data_t *anim_data4_ptr;
	int16_t dx2;
	int16_t dy2;
	int16_t draw_dx;
	int16_t draw_dy;
} anim_data_t;

typedef struct object_state_t {
	uint8_t displayed;
	int16_t life;
	int8_t anim_data_cycles;
	int8_t unk3;
	int16_t map_pos_x; /* init_map_pos_x ? */
	int16_t map_pos_y; /* init_map_pos_y ? */
	anim_data_t *anim_data1_ptr;
	anim_data_t *anim_data2_ptr;
	anim_data_t *anim_data3_ptr;
	anim_data_t *anim_data4_ptr;
	const uint8_t *move_data_ptr;
	int16_t distance_pos_x;
	int16_t distance_pos_y;
	int16_t pos_x;
	int16_t pos_y;
	uint16_t ref_ref_index;
	uint8_t unk26;
	uint8_t unk27;
	uint8_t transparent_flag;
	uint8_t unk29;
	uint8_t tile_num;
	uint8_t visible;
int16_t anim_data3_ptr_, anim_data4_ptr_, anim_data4_seg_;
	int16_t unk2C;
	int16_t unk2E;
} object_state_t;

typedef struct anim_frame_t {
	uint8_t frames_count;
	uint8_t action_op3_flag;
	int16_t frame_w;
	int16_t frame_h;
	int16_t _null1;
	int16_t frame_w2;
	int16_t delta_y;
	int16_t frame_h2;
	int16_t sprite_num;
	const struct anim_frame_t *next_anim_frame;
	int8_t draw_dy;
	int8_t draw_dx;
} anim_frame_t;

typedef struct {
	int16_t delta_x;
	int16_t delta_y;
	int16_t tile_map_offs_x;
	int16_t tile_map_offs_y;
	int16_t tile_map_end_offs_x;
	int16_t tile_map_end_offs_y;
	int16_t scroll_dx;
	int16_t scroll_dy;
	int16_t tile_block_x;
	int16_t tile_block_y;
	int16_t offs_x;
	int16_t offs_y;
	int16_t scroll_inc;
} decor_state_t;

typedef struct {
	uint16_t tile_num;
	uint16_t anim;
	uint16_t tiles_table[4];
} tile_anim_t;

typedef struct {
	int score;
	int extra_life_score;
	uint8_t boss_fight_counter;
	uint8_t boss_explosion_counter;
	int8_t tile_cycling_counter;
	tile_anim_t tile_anims_table[6];
} game_state_t;

#define MODE_WALKING 0
#define MODE_CLIMBING 1
#define MODE_JUMPING_FALLING 2

#define HDIR_FLAG_FACING_RIGHT   (1 << 0)
#define HDIR_FLAG_FALLING        (1 << 2)
#define HDIR_FLAG_FACING_CHANGED (1 << 3)
#define HDIR_FLAG_ON_STAIRS_UP   (1 << 4)
#define HDIR_FLAG_ON_STAIRS_DOWN (1 << 5)
#define HDIR_FLAG_CROUCHED       (1 << 6)

#define BIRD_DIR_FLAG_GOING_DOWN (1 << 2)

#define DECOR_REF_FLAG0_BLOCK_DOWN  (1 << 0)
#define DECOR_REF_FLAG0_BLOCK_UP    (1 << 1)
#define DECOR_REF_FLAG0_BLOCK_RIGHT (1 << 2)
#define DECOR_REF_FLAG0_BLOCK_LEFT  (1 << 3)

typedef struct {
	uint8_t unk0;
	uint8_t unk1; /* MODE_* */
	uint8_t hdir_flags; /* HDIR_FLAG_* */
	uint8_t vdir_flags;
	int16_t pos_x;
	int16_t pos_y;
	int16_t dim_w;
	int16_t dim_h;
	int16_t pos_dx_1;
	int16_t pos_dy_1;
	int16_t pos_dx_2;
	int16_t pos_dy_2;
	const uint16_t *move_offsets_data_ptr;
	int16_t anim_frames_count;
	uint8_t fire_button_counter;
	uint8_t throwing_ball_counter;
	uint8_t decor_ref_flag0; /* DECOR_REF_FLAG0_* */
	uint8_t fill_energy_bar;
	uint8_t lifes_count;
	uint8_t keys_count;
	uint8_t bullmodes_count;
	uint8_t hurricanes_count;
	uint8_t decor_ref_flag4; /* collision with decor */
	uint8_t decor_ref_flag3; /* interaction with decor (stairs etc.) */
	uint8_t dead_flag;
	uint8_t decor_ref_flag2; /* play sound */
	uint8_t has_blue_ball;
	int8_t throwing_apple_counter;
	uint8_t energy;
	uint8_t colliding;
	uint8_t action_code;
	int16_t bullmode_duration;
	int16_t has_wood;
	int16_t shield_duration;
	int16_t tile_blinking_duration;
	int16_t has_spring;
	int16_t tilemap_offset;
	uint16_t next_tilemap_offset;
	const anim_frame_t *anim_frame_ptr;
	uint8_t colliding_flag;
	int16_t stuck_duration; /* Carlos specific */
	int8_t power;
	int8_t increment_power_bar;
} player_state_t;

typedef struct {
	uint8_t unk0;
	uint8_t hdir_flags;
	int16_t pos_x;
	int16_t pos_y;
	int16_t dx;
	int16_t dy;
	int16_t frame_counter;
	int16_t frame_num;
} oscar_state_t;

typedef struct {
	int16_t pos_y;
	int16_t pos_x;
	uint16_t frame_counter;
	uint16_t frame_num;
	uint16_t last_frame_num;
	uint16_t frame_counter_mask;
} boss_state_t;

#endif // INTERN_H__
