
#include <getopt.h>
#ifdef __EMSCRIPTEN__
#include <emscripten.h>
#endif
#include "fileio.h"
#include "game.h"
#include "globals.h"
#include "input.h"
#include "resource.h"
#include "sound.h"
#include "systemstub.h"

#define STATE_PLAY_LEVEL 1
#define STATE_SHOW_IMAGE 2

static int _screen_counter;
static int _next_state, _state;

#define IMAGE_DURATION 100

static void do_image(bool key_released, int current_image) {
	if (current_image == IMAGE_MICROIDS) {
		if (key_released || _screen_counter >= IMAGE_DURATION) {
			g_game.screen_img = IMAGE_TITLE;
			snd_play_song(SNG_TITLE);
		}
	} else if (current_image == IMAGE_TITLE) {
		if (_screen_counter >= IMAGE_DURATION) {
			g_game.screen_img = IMAGE_CREDITS;
		} else if (key_released) {
			/* start level */
			g_game.screen_img = IMAGE_LEVELS + g_game.current_level;
		}
	} else if (current_image == IMAGE_CREDITS) {
		if (_screen_counter >= IMAGE_DURATION) {
			g_game.screen_img = IMAGE_TITLE;
		} else if (key_released) {
			/* start level */
			g_game.screen_img = IMAGE_LEVELS + g_game.current_level;
		}
	} else if (current_image >= IMAGE_LEVELS && current_image < IMAGE_LEVELS + GAME_LEVELS__v3) {
		if (key_released || _screen_counter >= IMAGE_DURATION) {
			/* next level */
			_next_state = STATE_PLAY_LEVEL;
		}
	} else if (current_image == IMAGE_LEVELS + GAME_LEVELS__v3) {
		/* game over */
		if (_screen_counter >= IMAGE_DURATION) {
			g_game.screen_img = IMAGE_TITLE;
			snd_play_song(SNG_TITLE);
			g_game.current_level = 0;
		}
	}
}

static void do_frame() {
	if (_next_state != _state) {
		_state = _next_state;
		if (_state == STATE_PLAY_LEVEL) {
			game_init_level(g_game.current_level);
		}
	}
	switch (_state) {
	case STATE_PLAY_LEVEL:
		game_run();
		if (g_game.quit_level_flag >= 2) {
			inp_end_level();
		} else if (g_game.new_level_flag) {
			inp_end_level();
			game_fini_level();
			++g_game.current_level;
			g_game.screen_img = IMAGE_LEVELS + g_game.current_level;
			_next_state = STATE_SHOW_IMAGE;
			if (g_game.current_level == GAME_LEVELS__v3) {
				snd_play_song(SNG_END);
			}
		}
		break;
	case STATE_SHOW_IMAGE:
		if (g_game.screen_img != g_res.img) {
			res_load_img(g_game.screen_img);
			g_game.screen_img = g_res.img;
			_screen_counter = 0;
		}
		do_image((sys_get_keys_released_mask() & SKM_ACTION) != 0, g_game.screen_img);
		++_screen_counter;
		break;
	}
	sys_update_screen();
	sys_clear_sprite_list();
}

int main(int argc, char *argv[]) {
	const char *data_path = 0;
	_screen_w = IS_GBA ? 240 : 320;
	_screen_h = 160;
	if (!IS_GBA && argc >= 2) {
		static struct option options[] = {
			{ "start_pos",   required_argument, 0, 1 },
			{ "level",       required_argument, 0, 2 },
			{ "screen_size", required_argument, 0, 3 },
			{ 0, 0, 0, 0 }
		};
		int index, c;

		data_path = strdup(argv[1]);
		while (1) {
			c = getopt_long(argc, argv, "", options, &index);
			if (c == -1) {
				break;
			}
			switch (c) {
			case 1:
				sscanf(optarg, "%d,%d", &g_options.start_pos_x, &g_options.start_pos_y);
				break;
			case 2:
				g_options.start_level = atoi(optarg);
				break;
			case 3:
				sscanf(optarg, "%d,%d", &_screen_w, &_screen_h);
				break;
			}
		}
	}
	fio_init(data_path ? data_path : "DATA");
	res_detect();
	sys_init(_screen_w, _screen_h, g_res.gfx_scale_w, g_res.gfx_scale_h, "Carlos");
	game_init();
	g_game.current_level = g_options.start_level;
	_state = -1;
	if (g_options.start_pos_x != 0 || g_options.start_pos_y != 0) {
		_next_state = STATE_PLAY_LEVEL;
	} else {
		g_game.screen_img = IMAGE_MICROIDS;
		_next_state = STATE_SHOW_IMAGE;
	}
#ifdef __EMSCRIPTEN__
        emscripten_set_main_loop(do_frame, 50, 0);
#else
	while (!sys_exit_flag) {
		do_frame();
		sys_process_events();
	}
	sys_fini();
#endif
	return 0;
}
