
#ifndef SYSTEMSTUB_H__
#define SYSTEMSTUB_H__

#include "util.h"

#define SKM_UP        (1 <<  0)
#define SKM_DOWN      (1 <<  1)
#define SKM_RIGHT     (1 <<  2)
#define SKM_LEFT      (1 <<  3)
#define SKM_ACTION    (1 <<  4)
#define SKM_QUIT      (1 <<  5)
#define SKM_PAUSE     (1 <<  6)
#define SKM_INVENTORY (1 <<  7)
#define SKM_F1        (1 <<  8)
#define SKM_F2        (1 <<  9)
#define SKM_F3        (1 << 10)
#define SKM_F4        (1 << 11)
#define SKM_F5        (1 << 12)
#define SKM_F6        (1 << 13)
#define SKM_F7        (1 << 14)
#define SKM_F8        (1 << 15)

#define SSF_VISIBLE  0x0000
#define SSF_HITMODE  0x8000
#define SSF_INVALID  0xFFFF

typedef enum {
	SSM_BITMAP,
	SSM_TILED
} sys_screen_mode_e;

typedef struct {
	int x, y;
	int w, h;
} sys_rect_t;

extern int sys_exit_flag;

extern void sys_init(int w, int h, int sw, int sy, const char *title);
extern void sys_fini();
extern void sys_set_screen_mode(sys_screen_mode_e mode);

extern void sys_set_palette_spr(const uint16_t *pal_data, int num_colors, int pal_num);
extern void sys_fade_in_palette();
extern void sys_fade_out_palette();

extern void sys_update_screen();
extern void sys_set_screen_offset(int offset);
extern void sys_fill_rect(const sys_rect_t *rect, int r, int g, int b);

extern void sys_set_bitmap_data(const uint8_t *bmp_data, int size);
extern void sys_set_tile_data(const uint8_t *tile_data, int size, uint8_t color);
extern void sys_set_tilemap_data(const uint8_t *map_data, int w, int h, const uint8_t *anim_lut);
extern void sys_set_tilemap_origin(int x, int y);
extern void sys_set_copper_pal(const uint16_t *pal_data, int num_lines);
extern void sys_set_background_frame(const uint8_t *data, int size);

extern void sys_clear_sprite_data();
extern int sys_load_sprite_data(const uint8_t *data, int size, int count, sys_rect_t *r);
extern void sys_get_sprite_dim(int spr_data, int num, int *w, int *h);
extern void sys_clear_sprite_list();
extern void sys_add_to_sprite_list(int spr_data, int num, int x, int y, int flags, int pal_num);

extern void sys_process_events();
extern int sys_get_keys_pressed_mask();
extern int sys_get_keys_released_mask();

extern void sys_play_sfx(int num, const uint8_t *data, int size);
extern void sys_play_module(int num, const uint8_t *data, int size);
extern void sys_stop_module();

extern void sys_print_string(const char *str);

#endif /* SYSTEMSTUB_H__ */
