
#include <time.h>
#include "util.h"
#include "systemstub.h"

uint16_t util_random_seed;
static int util_debug_mask = 0; /* DBG_FILEIO | DBG_GAME | DBG_RESOURCE; */

void string_upper(char *p) {
	for (; *p; ++p) {
		if (*p >= 'a' && *p <= 'z') {
			*p += 'A' - 'a';
		}
	}
}

void debug(int debug_channel, const char *msg, ...) {
	if (util_debug_mask & debug_channel) {
		char buf[256];
		va_list va;
		va_start(va, msg);
		vsnprintf(buf, sizeof(buf), msg, va);
		va_end(va);
		fprintf(stdout, "%s\n", buf);
		sys_print_string(buf);
	}
}

void warning(const char *msg, ...) {
	char buf[256];
	va_list va;
	va_start(va, msg);
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	fprintf(stdout, "WARNING: %s\n", buf);
	sys_print_string(buf);
}

void error(const char *msg, ...) {
	char buf[256];
	va_list va;
	va_start(va, msg);
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	fprintf(stdout, "ERROR: %s!\n", buf);
	sys_print_string(buf);
	exit(-1);
}

void info(const char *msg, ...) {
	char buf[256];
	va_list va;
	va_start(va, msg);
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	fprintf(stdout, "%s\n", buf);
	sys_print_string(buf);
}

void set_random_seed() {
	util_random_seed = (uint16_t)time(0);
}

uint16_t get_random_number(uint16_t mod) {
	uint16_t r1 = util_random_seed;
	uint16_t r2 = r1;
	assert(mod != 0);
	r1 <<= 1;
	r1 += r2;
	r1 <<= 1;
	r1 += r2 * 3;
	r1 <<= 2;
	r1 += r2;
	r1 <<= 2;
	r1 += r2;
	r1 <<= 2;
	r1 += r2;
	r1 <<= 1;
	r1 += r2;
	r1 <<= 2;
	r1 += r2;
	r1 <<= 2;
	r1 += r2;
	util_random_seed = r1 + 1;
	return util_random_seed % mod;
}

