
#ifndef UTIL_H__
#define UTIL_H__

#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

#define DBG_GAME      (1 << 0)
#define DBG_SYSTEM    (1 << 1)
#define DBG_FILEIO    (1 << 2)
#define DBG_RESOURCE  (1 << 3)

#define GCC_PACK __attribute__((packed))
#define GCC_ALIGN(x) __attribute__ ((__aligned__ (x)))
#define GCC_INLINE __inline__

#ifdef CARLOS_GBA_VERSION
#define GCC_EXT __attribute__((section(".ewram")))
#define IS_GBA true
#else
#define GCC_EXT
#define IS_GBA false
#endif

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#ifndef ABS
#define ABS(a) (((a)<0)?(-(a)):(a))
#endif
#ifndef ARRAYSIZE
#define ARRAYSIZE(a) (sizeof(a)/sizeof(a[0]))
#endif

extern void string_upper(char *p);
extern void debug(int debug_channel, const char *msg, ...);
extern void warning(const char *msg, ...);
extern void error(const char *msg, ...);
extern void info(const char *msg, ...);

static GCC_INLINE uint16_t READ_LE_UINT16(const void *ptr) {
	if (IS_GBA) {
		if ((((uintptr_t)ptr) & 1) == 0) {
			return *(uint16_t *)ptr;
		}
	}
	const uint8_t *b = (const uint8_t *)ptr;
	return (b[1] << 8) | b[0];
}

static GCC_INLINE uint32_t READ_LE_UINT32(const void *ptr) {
	if (IS_GBA) {
		if ((((uintptr_t)ptr) & 3) == 0) {
			return *(uint32_t *)ptr;
		}
	}
	const uint8_t *b = (const uint8_t *)ptr;
	return (b[3] << 24) | (b[2] << 16) | (b[1] << 8) | b[0];
}

static GCC_INLINE uint16_t READ_BE_UINT16(const void *ptr) {
	const uint8_t *b = (const uint8_t *)ptr;
	return (b[0] << 8) | b[1];
}

static GCC_INLINE uint32_t READ_BE_UINT32(const void *ptr) {
	const uint8_t *b = (const uint8_t *)ptr;
	return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];
}

extern uint16_t util_random_seed;

void set_random_seed();
uint16_t get_random_number(uint16_t mode);

#endif // UTIL_H__
